# -*- coding: utf-8 -*-
"""
    { Types } for GraphQL
"""

from typing import Optional
import fastberry as fb

# Create your <types> here.
@fb.sql.model
class Task:
    """Task's Database Table"""

    title: str
    description: str
    status: str



"""
    On Events (FastAPI)
"""
from fastapi.responses import RedirectResponse

from ..handlers.cmd.action import introspection_info


def redirect_root(context):
    """Redirect"""

    @context.app.get("/", response_class=RedirectResponse)
    async def redirect_to_docs():
        """Redirect To Docs"""
        return "/docs"

    return context


def graphql_info(context):
    """Get Graphql INFO"""
    schema = context.controller.graphql.schema()

    @context.app.get("/graphql-info")
    async def get_graphql_info():
        """Get Graphql INFO"""
        return introspection_info(schema)

    return context



"""
    Extras [middleware, on_startup, on_shutdown, extensions, permissions]
"""
import functools

from ..core import types


def add_middleware(app=None, installed: list = None):
    """Install The Middleware"""
    for middleware in reversed(installed):
        app.add_middleware(middleware)
    return app


def extras(items):
    """Collect Extras From (`spoc.toml`)"""
    extra_out = dict(items)
    extra_out["middleware"] = functools.partial(
        add_middleware, installed=extra_out.get("middleware", {})
    )
    return types.Extras(**extra_out)



# -*- coding: utf-8 -*-
"""
    { Forms } for GraphQL
"""
import fastberry as fb

# Create Group "Form"
form = fb.input("form")

# Create your <forms> here.
@form
class Task:
    """Task's Client Input Form"""

    title = fb.value(str, default=None, required=True)
    description = fb.value(str, default=None, required=True)
    status = fb.value(str, default=None, required=True)



# -*- coding: utf-8 -*-
"""
    { Controller } for the Database(s)
"""

import fastberry as fb

from . import types

# Create your <managers> here.
@fb.manager
class Task:
    """Task's Database Manager"""

    model = types.Task

    @classmethod
    async def create(cls, form):
        """Create"""
        # Errors
        errors_messages = []

        # Good Input
        if form.is_valid:
            results = await cls.objects.create(
                {
                    "title": form.data.title,
                    "status": form.data.status,
                    "description": form.data.description,
                }
            )
            if not results.error:
                item = results.data
                return types.Task(
                    _id=item._id,
                    id=item.id,
                    title=item.title,
                    status=item.status,
                    description=item.description,
                )

        # Bad Input
        errors_messages.append(fb.error(type="input", text="Something Went Wrong!"))
        return fb.errors(messages=errors_messages)

    @classmethod
    async def update(cls, unique_id, status):
        """Update"""
        # Errors
        errors_messages = []

        # Good Input
        form = {
            "status": status,
        }
        results = await cls.objects.update(unique_id, form)
        if not results.error:
            item = results.data
            return types.Task(
                _id=item._id,
                id=item.id,
                title=item.title,
                status=item.status,
                description=item.description,
            )

        # Bad Input
        errors_messages.append(fb.error(type="input", text="Something Went Wrong!"))
        return fb.errors(messages=errors_messages)

    @classmethod
    async def delete(cls, unique_id):
        """Delete"""
        results = await cls.objects.delete(unique_id)
        if not results.error:
            return True
        return False

    @classmethod
    async def detail(cls, unique_id):
        """Detail"""
        results = await cls.objects.detail(unique_id)
        return results

    @classmethod
    async def search(cls, search=None):
        """Search"""
        columns = ["status"]
        results = await cls.objects.search(
            columns=columns, value=search, page=1, limit=100, sort_by="-id"
        )
        # Search
        if not results.error:
            return fb.page(
                edges=[
                    types.Task(
                        _id=item._id,
                        id=item.id,
                        title=item.title,
                        status=item.status,
                        description=item.description,
                    )
                    for item in results.data
                ],
                length=results.count,
                pages=results.pages,
            )
        # All Items
        return fb.page(
            edges=[],
            length=0,
            pages=0,
        )



"""
    GraphQL - Strawberry
"""
import functools
import typing
from collections import namedtuple

import spoc
import strawberry
from strawberry.types import Info

from ...components import COMPONENT
from .schema import Schema

GraphQL = namedtuple("GraphQL", ["schema", "info", "operations"])
GraphQLInfo = namedtuple("Info", ["query", "mutation"])
Operation = namedtuple("Operation", ["name", "annotations"])


def operation_annotations(annotations: dict):
    """GraphQL Operations"""
    ops_annotations = {"inputs": {}}
    for key, val in annotations:
        if key == "return":
            return_types = set()
            for item in typing.get_args(val):
                if hasattr(item, "__name__"):  # hasattr(item, "__spoc__") and
                    return_types.add(item.__name__)
            ops_annotations["return"] = list(return_types)
        else:
            if hasattr(val, "__spoc__") and hasattr(val, "__name__"):
                ops_annotations["inputs"][key] = val.__name__
    return ops_annotations


def create_operation(name, annotations):
    """Create Operations NamedTuple"""
    return Operation(
        name=name,
        annotations=operation_annotations(annotations.items()),
    )


def create_name(method: object, app: str = None, model: str = None):
    """GraphQL Resolver's Name"""
    if model:
        method_name = f"{model}_{method.__name__}"
    else:
        method_name = method.__name__
    if app:
        method_name = f"{app}_{method_name}"
    return method_name


def get_module_name(model: object):
    """Get: Class-Module's Name"""
    name = model.__module__
    parts = name.split(".")
    if parts[0] == "apps" and len(parts) > 1:
        parts.pop(0)
    module_name = parts[0]
    if module_name == "__main__":
        module_name = "main"
    return module_name


def graphql(schemas: list = None, permissions: list = None):
    """Collect (GraphQL) Strawberry"""
    gql_schema = {"Query": {}, "Mutation": {}}
    permission_classes = permissions or []
    fastberry_operations = []
    # Collect
    for active in schemas:
        is_component = spoc.is_component(active.object, COMPONENT["graphql"])
        if is_component:
            cls_fields = spoc.get_fields(active.object)
            method_model = None
            method_app = get_module_name(active.object)
            if hasattr(active.object, "Meta"):
                if hasattr(active.object.Meta, "model"):
                    method_model = getattr(active.object.Meta, "model")
                if hasattr(active.object.Meta, "app"):
                    method_app = getattr(active.object.Meta, "app")
            if "Query" in cls_fields:
                cls_query = getattr(active.object, "Query")
                cls_query_fields = spoc.get_fields(cls_query)
                for current in cls_query_fields:
                    resolver_func = getattr(cls_query, current)
                    resolver_name = create_name(
                        resolver_func, app=method_app, model=method_model
                    )
                    op = functools.partial(
                        create_operation, resolver_name, resolver_func.__annotations__
                    )
                    fastberry_operations.append(op)
                    resolver_func.__annotations__["info"] = Info
                    gql_schema["Query"][resolver_name] = strawberry.field(
                        resolver=resolver_func,
                        description=(resolver_func.__doc__ or "").strip(),
                        permission_classes=permission_classes,
                    )
            if "Mutation" in cls_fields:
                cls_query = getattr(active.object, "Mutation")
                cls_query_fields = spoc.get_fields(cls_query)
                for current in cls_query_fields:
                    resolver_func = getattr(cls_query, current)
                    resolver_name = create_name(
                        resolver_func, app=method_app, model=method_model
                    )
                    op = functools.partial(
                        create_operation, resolver_name, resolver_func.__annotations__
                    )
                    fastberry_operations.append(op)
                    resolver_func.__annotations__["info"] = Info
                    gql_schema["Mutation"][resolver_name] = strawberry.field(
                        resolver=resolver_func,
                        description=(resolver_func.__doc__ or "").strip(),
                        permission_classes=permission_classes,
                    )
    # Build Schema
    Query = None
    Mutation = None
    if gql_schema["Query"]:
        Query = type("Query", (object,), gql_schema["Query"])
        Query = strawberry.type(Query)
    if gql_schema["Mutation"]:
        Mutation = type("Mutation", (object,), gql_schema["Mutation"])
        Mutation = strawberry.type(Mutation)
    return GraphQL(
        operations=fastberry_operations,
        schema=functools.partial(Schema, query=Query, mutation=Mutation),
        info=GraphQLInfo(
            query=frozenset(gql_schema["Query"].keys()),
            mutation=frozenset(gql_schema["Mutation"].keys()),
        ),
    )
